@@ -1,7 +1,5 @@
-# $Id: Client.pm,v 1.12 2006/04/23 08:37:41 joern Exp $
-
 #-----------------------------------------------------------------------
-# Copyright (C) 2002-2006 Jrn Reder <joern AT zyn.de>.
+# Copyright (C) 2005-2015 by Jörn Reder <joern AT zyn.de>.
 # All Rights Reserved. See file COPYRIGHT for details.
 #
 # This module is part of Event::RPC, which is free software; you can
@@ -11,55 +9,84 @@
 package Event::RPC::Client;
 
 use Event::RPC;
-use Event::RPC::Message;
+use Event::RPC::Message::Negotiate;
 
 use Carp;
 use strict;
+use utf8;
+
 use IO::Socket::INET;
 
+#-- This can be changed for testing purposes e.g. to simulate
+#-- old clients connecting straight with Storable format.
+$Event::RPC::Client::DEFAULT_MESSAGE_FORMAT = "Event::RPC::Message::Negotiate";
+
 sub get_client_version          { $Event::RPC::VERSION                  }
 sub get_client_protocol         { $Event::RPC::PROTOCOL                 }
 
 sub get_host                    { shift->{host}                         }
 sub get_port                    { shift->{port}                         }
 sub get_sock                    { shift->{sock}                         }
+sub get_timeout                 { shift->{timeout}                      }
 sub get_classes                 { shift->{classes}                      }
 sub get_class_map               { shift->{class_map}                    }
 sub get_loaded_classes          { shift->{loaded_classes}               }
 sub get_error_cb                { shift->{error_cb}                     }
 sub get_ssl                     { shift->{ssl}                          }
+sub get_ssl_ca_file             { shift->{ssl_ca_file}                  }
+sub get_ssl_ca_path             { shift->{ssl_ca_path}                  }
+sub get_ssl_opts                { shift->{ssl_opts}                     }
 sub get_auth_user               { shift->{auth_user}                    }
 sub get_auth_pass               { shift->{auth_pass}                    }
 sub get_connected               { shift->{connected}                    }
 sub get_server                  { shift->{server}                       }
 sub get_server_version          { shift->{server_version}               }
 sub get_server_protocol         { shift->{server_protocol}              }
+sub get_message_format          { shift->{message_format}               }
+sub get_insecure_msg_fmt_ok     { shift->{insecure_msg_fmt_ok}          }
 
 sub set_host                    { shift->{host}                 = $_[1] }
 sub set_port                    { shift->{port}                 = $_[1] }
 sub set_sock                    { shift->{sock}                 = $_[1] }
+sub set_timeout                 { shift->{timeout}              = $_[1] }
 sub set_classes                 { shift->{classes}              = $_[1] }
 sub set_class_map               { shift->{class_map}            = $_[1] }
 sub set_loaded_classes          { shift->{loaded_classes}       = $_[1] }
 sub set_error_cb                { shift->{error_cb}             = $_[1] }
 sub set_ssl                     { shift->{ssl}                  = $_[1] }
+sub set_ssl_ca_file             { shift->{ssl_ca_file}          = $_[1] }
+sub set_ssl_ca_path             { shift->{ssl_ca_path}          = $_[1] }
+sub set_ssl_opts                { shift->{ssl_opts}             = $_[1] }
 sub set_auth_user               { shift->{auth_user}            = $_[1] }
 sub set_auth_pass               { shift->{auth_pass}            = $_[1] }
 sub set_connected               { shift->{connected}            = $_[1] }
 sub set_server                  { shift->{server}               = $_[1] }
 sub set_server_version          { shift->{server_version}       = $_[1] }
 sub set_server_protocol         { shift->{server_protocol}      = $_[1] }
+sub set_message_format          { shift->{message_format}       = $_[1] }
+sub set_insecure_msg_fmt_ok     { shift->{insecure_msg_fmt_ok}  = $_[1] }
+
+sub get_max_packet_size {
+    return Event::RPC::Message->get_max_packet_size;
+}
+
+sub set_max_packet_size {
+    my $class = shift;
+    my ($value) = @_;
+    Event::RPC::Message->set_max_packet_size($value);
+}
 
 sub new {
     my $class = shift;
     my %par   = @_;
-    my  ($server, $host, $port, $classes, $class_map, $error_cb) =
-    @par{'server','host','port','classes','class_map','error_cb'};
-    my  ($ssl, $auth_user, $auth_pass) =
-    @par{'ssl','auth_user','auth_pass'};
+    my  ($server, $host, $port, $classes, $class_map, $error_cb, $timeout) =
+    @par{'server','host','port','classes','class_map','error_cb','timeout'};
+    my  ($ssl, $ssl_ca_file, $ssl_opts, $auth_user, $auth_pass, $insecure_msg_fmt_ok) =
+    @par{'ssl','ssl_ca_file','ssl_opts','auth_user','auth_pass','insecure_msg_fmt_ok'};
 
     $server ||= '';
     $host   ||= '';
+    $insecure_msg_fmt_ok = 1 unless defined $insecure_msg_fmt_ok;
 
     if ( $server ne '' and $host eq '' ) {
         warn "Option 'server' is deprecated. Use 'host' instead.";
@@ -67,17 +94,22 @@ sub new {
     }
 
     my $self = bless {
-        host           => $server,
-        server         => $host,
-        port           => $port,
-        classes        => $classes,
-        class_map      => $class_map,
-        ssl            => $ssl,
-        auth_user      => $auth_user,
-        auth_pass      => $auth_pass,
-        error_cb       => $error_cb,
-        loaded_classes => {},
-        connected      => 0,
+        host                => $server,
+        server              => $host,
+        port                => $port,
+        timeout             => $timeout,
+        classes             => $classes,
+        class_map           => $class_map,
+        ssl                 => $ssl,
+        ssl_ca_file         => $ssl_ca_file,
+        ssl_opts            => $ssl_opts,
+        auth_user           => $auth_user,
+        auth_pass           => $auth_pass,
+        error_cb            => $error_cb,
+        message_format      => $Event::RPC::Client::DEFAULT_MESSAGE_FORMAT,
+        insecure_msg_fmt_ok => $insecure_msg_fmt_ok,
+        loaded_classes      => {},
+        connected           => 0,
     }, $class;
 
     return $self;
@@ -88,46 +120,85 @@ sub connect {
 
     croak "Client is already connected" if $self->get_connected;
 
-    my $ssl    = $self->get_ssl;
-    my $server = $self->get_server;
-    my $port   = $self->get_port;
+    my $ssl     = $self->get_ssl;
+    my $server  = $self->get_server;
+    my $port    = $self->get_port;
+    my $timeout = $self->get_timeout;
 
-    if ($ssl) {
+    $self->set_message_format($Event::RPC::Client::DEFAULT_MESSAGE_FORMAT);
+
+    #-- Client may try to fallback to Storable
+    Event::RPC::Message::Negotiate->set_storable_fallback_ok(1)
+        if $self->get_message_format eq 'Event::RPC::Message::Negotiate' and
+           $self->get_insecure_msg_fmt_ok;
+
+    if ( $ssl ) {
         eval { require IO::Socket::SSL };
         croak "SSL requested, but IO::Socket::SSL not installed" if $@;
     }
 
     my $sock;
-    if ($ssl) {
+    if ( $ssl ) {
+        my @verify_opts;
+        if ( $self->get_ssl_ca_file or $self->get_ssl_ca_path ) {
+            push @verify_opts, (
+                SSL_verify_mode => IO::Socket::SSL::SSL_VERIFY_PEER(),
+                SSL_ca_file     => $self->get_ssl_ca_file,
+                SSL_ca_path     => $self->get_ssl_ca_path,
+            );
+        }
+        else {
+            push @verify_opts, (
+                SSL_verify_mode => IO::Socket::SSL::SSL_VERIFY_NONE(),
+            );
+        }
+
+        my $ssl_opts = $self->get_ssl_opts;
+
         $sock = IO::Socket::SSL->new(
             Proto    => 'tcp',
             PeerPort => $port,
             PeerAddr => $server,
-            Type     => SOCK_STREAM
-            )
-            or croak
-            "Can't open SSL connection to $server:$port: $IO::Socket::SSL::ERROR";
+            Type     => SOCK_STREAM,
+            Timeout  => $timeout,
+            @verify_opts,
+            ($ssl_opts?%{$ssl_opts}:()),
+        )
+        or croak "Can't open SSL connection to $server:$port: $IO::Socket::SSL::ERROR";
     }
     else {
         $sock = IO::Socket::INET->new(
             Proto    => 'tcp',
             PeerPort => $port,
             PeerAddr => $server,
-            Type     => SOCK_STREAM
-            )
-            or croak "Can't open connection to $server:$port - $!";
+            Type     => SOCK_STREAM,
+            Timeout  => $timeout,
+        )
+        or croak "Can't open connection to $server:$port - $!";
     }
 
     $sock->autoflush(1);
 
     $self->set_sock($sock);
 
-    $self->check_version;
+    eval {
+        #-- Perform message format negotitation if we are not
+        #-- configured to a specific format already.
+        $self->negotiate_message_format
+            if $self->get_message_format eq 'Event::RPC::Message::Negotiate';
+
+        $self->check_version;
+    };
+
+    if ( $@ ) {
+        $self->disconnect;
+        die $@;
+    }
 
     my $auth_user = $self->get_auth_user;
     my $auth_pass = $self->get_auth_pass;
 
-    if ($auth_user) {
+    if ( $auth_user ) {
         my $rc = $self->send_request(
             {   cmd  => 'auth',
                 user => $auth_user,
@@ -162,8 +233,8 @@ sub log_connect {
         PeerPort => $port,
         PeerAddr => $server,
         Type     => SOCK_STREAM
-        )
-        or croak "Can't open connection to $server:$port - $!";
+    )
+    or croak "Can't open connection to $server:$port - $!";
 
     return $sock;
 }
@@ -197,10 +268,70 @@ sub error {
     1;
 }
 
+sub negotiate_message_format {
+    my $self = shift;
+
+    my $rc = eval {
+        $self->send_request({
+            cmd => "neg_formats_avail"
+        })
+    };
+
+    if ( $@ ) {
+        #-- On error we probably may fall back to Storable
+        #-- (we connected to an old server)
+        if ( $self->get_insecure_msg_fmt_ok ) {
+            require Event::RPC::Message::Storable;
+            $self->set_message_format("Event::RPC::Message::Storable");
+            return;
+        }
+
+        #-- die if Storable is not allowed
+        die "Error on message format negotiation and client is not ".
+            "allowed to fall back to Storable\n";
+    }
+
+    my $modules_by_format_name =
+        Event::RPC::Message::Negotiate->known_message_formats;
+
+    my @formats = split(/,/, $rc->{msg});
+
+    my $format_chosen = '';
+    my $module_chosen = '';
+    foreach my $format ( @formats ) {
+        my $module = $modules_by_format_name->{$format}
+            or die "Unknown message format '$format";
+
+        eval "use $module";
+
+        if ( not $@ ) {
+            $format_chosen = $format;
+            $module_chosen = $module;
+            last;
+        };
+    }
+
+    die "Can't negotiate message format\n" unless $format_chosen;
+
+    eval {
+        $self->send_request({
+            cmd => "neg_format_set",
+            msg => $format_chosen,
+        })
+    };
+
+    die "Error on neg_format_set: $@" if $@;
+
+    $self->set_message_format($module_chosen);
+
+    1;
+}
+
 sub check_version {
     my $self = shift;
 
-    my $rc = $self->send_request( { cmd => 'version', } );
+    my $rc = eval { $self->send_request( { cmd => 'version', } ) };
+    die "CATCHED $@" if $@;
 
     $self->set_server_version( $rc->{version} );
     $self->set_server_protocol( $rc->{protocol} );
@@ -389,7 +520,7 @@ sub send_request {
     my $self = shift;
     my ($request) = @_;
 
-    my $message = Event::RPC::Message->new( $self->get_sock );
+    my $message = $self->get_message_format->new( $self->get_sock );
 
     $message->write_blocked($request);
 
@@ -412,6 +543,8 @@ sub send_request {
 
 __END__
 
+=encoding utf8
+
 =head1 NAME
 
 Event::RPC::Client - Client API to connect to Event::RPC Servers
@@ -429,11 +562,18 @@ Event::RPC::Client - Client API to connect to Event::RPC Servers
     classes   => [ "Event::RPC::Test" ],
     class_map => { "Event::RPC::Test" => "My::Event::RPC::Test" },
 
-    ssl       => 1,
+    ssl         => 1,
+    ssl_ca_file => "some/ca.crt",
+    ssl_ca_path => "some/ca/dir",
+    ssl_opts    => { SSL_verifycn_name => 'server-hostname' },
+
+    timeout     => 10,
 
     auth_user => "fred",
     auth_pass => Event::RPC->crypt("fred",$password),
 
+    insecure_msg_fmt_ok => 1,
+
     error_cb => sub {
       my ($client, $error) = @_;
       print "An RPC error occured: $error\n";
@@ -442,10 +582,11 @@ Event::RPC::Client - Client API to connect to Event::RPC Servers
     },
   );
 
+  $rpc_client->set_max_packet_size(2*1024*1024*1024);
   $rpc_client->connect;
   
-  #-- And now use classes and methods the server
-  #-- allows to access via RPC, here My::TestModule
+  #-- And now use classes and methods to which the
+  #-- server allows access via RPC, here My::TestModule
   #-- from the Event::RPC::Server manpage SYNPOSIS.
   my $obj = My::TestModule->new( data => "foobar" );
   print "obj says hello: ".$obj->hello."\n";
@@ -499,6 +640,17 @@ This is the TCP port the server is listening to.
 
 =back
 
+=head2 NETWORK OPTIONS
+
+=over 4
+
+=item B<timeout>
+
+Specify a timeout (in seconds), which is applied when connecting
+the server.
+
+=back
+
 =head2 CLASS IMPORT OPTION
 
 =over 4
@@ -545,10 +697,13 @@ still write this on the server to create objects there:
 
 =back
 
-=head2 SSL OPTION
+=head2 SSL OPTIONS
 
 If the server accepts only SSL connections you need to enable
-ssl here in the client as well:
+ssl here in the client as well. By default the SSL connection
+will be established without any peer verification, which makes
+Man-in-the-Middle attacks possible. If you want to prevent that,
+you need to set either B<ssl_ca_file> or B<ssl_ca_path> option.
 
 =over 4
 
@@ -556,6 +711,27 @@ ssl here in the client as well:
 
 Set this option to 1 to encrypt the network connection using SSL.
 
+=item B<ssl_ca_file>
+
+Path to the the Certificate Authority's certificate file
+(ca.crt), your server key was signed with.
+
+=item B<ssl_ca_path>
+
+Path of a directory containing several trusted certificates with
+a proper index. Please refer to the OpenSSL documentation for
+details about setting up such a directory.
+
+=item B<ssl_opts>
+
+This optional parameter takes a hash reference of options
+passed to IO::Socket::SSL->new(...) to have more control
+over the SSL connection. For example you can set the
+'SSL_verifycn_name' here if the server certificate common
+name doesn't match to the hostname you use to resolve
+the server IP or use you have to use a static server IP
+address or something like that. 
+
 =back
 
 =head2 AUTHENTICATION OPTIONS
@@ -586,6 +762,36 @@ use this method:
 If the passed credentials are invalid the Event::RPC::Client->connect()
 method throws a correspondent exception.
 
+=head2 MESSAGE FORMAT OPTIONS
+
+Event::RPC supports different CPAN modules for data serialisation,
+named "message formats" here:
+
+  SERL -- Sereal::Encoder, Sereal::Decoder
+  CBOR -- CBOR::XS
+  JSON -- JSON::XS
+  STOR -- Storable
+
+Server and client negotiate automatically which format is
+best to use. The server sends a list of supported formats
+to the client which takes the first one which is available.
+
+For the client there is one option to influence this
+format negotiation mechanism:
+
+=over 4
+
+=item B<insecure_msg_fmt_ok>
+
+The Storable module is known to be insecure, so it should
+be taken as the last option only. By default the Client
+would do so. You can prevent that by setting this option
+explicitely to 0. It's enabled by default. Most likely
+the connection will fail in that case, because the server
+only will offer Storable if no other serialiser is available.
+
+=back
+
 =head2 ERROR HANDLING
 
 Any exceptions thrown on the server during execution of a remote
@@ -631,6 +837,22 @@ Closes the connection to the server. You may omit explicit disconnecting
 since it's done automatically once the Event::RPC::Client object gets
 destroyed.
 
+=item $rpc_client->B<set_max_packet_size> ( $bytes )
+
+By default Event::RPC does not handle network packages which
+exceed 2 GB in size (was 4 MB with version 1.04 and earlier).
+
+You can change this value using this method at any time,
+but 4 GB is the maximum. An attempt of the server to send a
+bigger packet will be aborted and reported as an exception
+on the client and logged as an error message on the server.
+
+Note: you have to set the same value on client and server side!
+
+=item $rpc_client->B<get_max_packet_size>
+
+Returns the currently active max packet size.
+
 =back
 
 =head1 READY ONLY ATTRIBUTES
@@ -657,11 +879,11 @@ Returns the Event::RPC protocol number of the client.
 
 =head1 AUTHORS
 
-  Jrn Reder <joern at zyn dot de>
+  Jörn Reder <joern AT zyn.de>
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2002-2006 by Joern Reder, All Rights Reserved.
+Copyright (C) 2005-2015 by Jörn Reder <joern AT zyn.de>.
 
 This library is free software; you can redistribute it
 and/or modify it under the same terms as Perl itself.
